/*
 * Copyright 2019 Bloomberg Finance LP
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef INCLUDED_BUILDBOXWORKER_EXPIRETIME
#define INCLUDED_BUILDBOXWORKER_EXPIRETIME

#include <google/protobuf/timestamp.pb.h>

#include <chrono>

namespace buildboxworker {

constexpr float TIMEOUT_PADDING_FACTOR = 0.20;

struct ExpireTime {
    static std::chrono::system_clock::time_point
    convertToTimePoint(const google::protobuf::Timestamp &timestamp);

    static constexpr float updateTimeoutPaddingFactor()
    {
        return TIMEOUT_PADDING_FACTOR;
    };
};

} // namespace buildboxworker

#endif
