// Copyright 2021 Bloomberg Finance L.P
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// component under test
#include <rumbad_queue.h>

// third party includes
#include <gmock/gmock.h>
#include <gtest/gtest.h>

namespace rumbad {
namespace test {

using namespace ::testing;

TEST(QueuePushPop, AsExpected)
{
    Queue<std::string> testQueue;
    std::string testString = "hello world";

    ASSERT_EQ(testQueue.size(), 0);

    testQueue.push(testString);
    ASSERT_EQ(testQueue.size(), 1);

    std::string poppedValue;
    bool popped = false;
    std::tie(poppedValue, popped) = testQueue.pop();
    ASSERT_EQ(testQueue.size(), 0);
    ASSERT_TRUE(popped);
    ASSERT_EQ(testString, poppedValue);

    std::tie(poppedValue, popped) = testQueue.pop();
    ASSERT_FALSE(popped);
}

} // namespace test
} // namespace rumbad
