// Copyright 2020 Bloomberg Finance L.P
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <buildboxcommonmetrics_countingmetricvalue.h>
#include <gtest/gtest.h>

using namespace buildboxcommon::buildboxcommonmetrics;

TEST(MetricsTest, CountingMetricValueConstructorSetGet)
{
    CountingMetricValue myValue(42);
    EXPECT_EQ(myValue.value(), 42);
    myValue.setValue(43);
    EXPECT_EQ(myValue.value(), 43);
}

TEST(MetricsTest, CountingMetricValueSetGet)
{
    CountingMetricValue myValue;
    EXPECT_EQ(myValue.value(), 0);
    myValue.setValue(4);
    EXPECT_EQ(myValue.value(), 4);
    EXPECT_EQ(myValue.toStatsD("foo"), "foo:4|c");
}

TEST(MetricsTest, CountingMetricValueAdd)
{
    CountingMetricValue myValue(42);
    CountingMetricValue anotherValue(10);

    myValue += anotherValue;
    EXPECT_EQ(myValue.value(), 52);
}

TEST(MetricsTest, CountingMetricValueIncrement)
{
    CountingMetricValue myValue(42);
    myValue++;
    EXPECT_EQ(myValue.value(), 43);
}

TEST(MetricsTest, CountingMetricValueEquality)
{
    EXPECT_TRUE(CountingMetricValue(2) == CountingMetricValue(2));
    EXPECT_FALSE(CountingMetricValue(2) != CountingMetricValue(2));
}

TEST(MetricsTest, CountingMetricValueInequality)
{
    EXPECT_FALSE(CountingMetricValue(2) == CountingMetricValue(3));
    EXPECT_TRUE(CountingMetricValue(2) != CountingMetricValue(3));
}
