// Copyright 2020 Bloomberg Finance L.P
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef BUILDBOXCOMMONMETRICS_STATSDPUBLISHERCREATOR_H
#define BUILDBOXCOMMONMETRICS_STATSDPUBLISHERCREATOR_H

#include <memory>

#include <buildboxcommonmetrics_metricsconfigurator.h>
#include <buildboxcommonmetrics_statsdpublisher.h>

namespace buildboxcommon {

namespace buildboxcommonmetrics {

using StatsDPublisherType =
    MetricsConfigurator::publisherTypeOfAllValueTypes<StatsDPublisher>;

class StatsdPublisherCreator {
  public:
    // Return a publisher type of either stderr, udp server or file.
    //
    // If 'metricsConfig' specifies a udpServer, the parsing of the port
    // may throw std::invalid_argument or std::range_error exceptions
    static std::shared_ptr<StatsDPublisherType>
    createStatsdPublisher(const MetricsConfigType &metricsConfig)
    {
        return MetricsConfigurator::createMetricsPublisherWithConfig<
            StatsDPublisherType>(metricsConfig);
    }

    StatsdPublisherCreator() = delete;
};

} // namespace buildboxcommonmetrics
} // namespace buildboxcommon
#endif
