/*
 * Copyright 2024 Bloomberg LP
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef INCLUDED_BUILDBOXCOMMON_CONNECTIONOPTIONS_TOML
#define INCLUDED_BUILDBOXCOMMON_CONNECTIONOPTIONS_TOML

#include <toml++/toml.h>

#include <buildboxcommon_commandline.h>
#include <buildboxcommon_connectionoptions.h>

namespace buildboxcommon {

class ConnectionOptionsTOML {
  public:
    static ConnectionOptions configureChannel(const toml::table &tomlConfig);

    static void updateChannel(const toml::table &tomlConfig,
                              ConnectionOptions &channel);
};

} // namespace buildboxcommon

#endif
