// Copyright 2019 Bloomberg Finance L.P
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <buildboxcommonmetrics_durationmetricvalue.h>

namespace buildboxcommon {
namespace buildboxcommonmetrics {

DurationMetricValue::DurationMetricValue(TimeDenomination value)
    : d_value(value)
{
}

void DurationMetricValue::setValue(DurationMetricValue::TimeDenomination value)
{
    d_value = value;
}
DurationMetricValue::TimeDenomination DurationMetricValue::value() const
{
    return d_value;
}
const std::string
DurationMetricValue::toStatsD(const std::string &myName) const
{
    return std::string(
        myName + ":" +
        std::to_string(
            std::chrono::duration_cast<std::chrono::milliseconds>(value())
                .count()) +
        "|ms");
}

bool DurationMetricValue::operator==(const DurationMetricValue &other) const
{
    return this->value() == other.value();
}

bool DurationMetricValue::operator!=(const DurationMetricValue &other) const
{
    return !(*this == other);
}

} // namespace buildboxcommonmetrics
} // namespace buildboxcommon
