// Copyright (C) 2025 Bloomberg LP
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//  <http://www.apache.org/licenses/LICENSE-2.0>
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto3";

package build.buildgrid;

// SchedulingMetadata is used to provide metadata to schedule remote
// executions, optionally attached to each `ExecuteRequest`.
// Each field is expected to change or hint where the job is scheduled,
// but behavior of the job itself cannot depend on it.
message SchedulingMetadata {
  string locality_hint = 1;
}
