// Copyright 2020 Bloomberg Finance L.P
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <buildboxcommonmetrics_distributionmetricvalue.h>
#include <gtest/gtest.h>

using namespace buildboxcommon::buildboxcommonmetrics;

TEST(MetricsTest, DistributionMetricValueAbsoluteByDefault)
{
    const DistributionMetricValue v(1024);
    EXPECT_EQ(v.value(), 1024);
}

TEST(MetricsTest, DistributionMetricStatsD)
{
    const DistributionMetricValue v(100);
    EXPECT_EQ(v.toStatsD("distribution-name"), "distribution-name:100|d");
}

TEST(MetricsTest, DistributionMetricValueComparison)
{
    ASSERT_EQ(DistributionMetricValue(123), DistributionMetricValue(123));
    ASSERT_NE(DistributionMetricValue(123), DistributionMetricValue(456));
}
