// Copyright 2020 Bloomberg Finance L.P
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef BUILDBOXWORKER_METRICNAMES_H
#define BUILDBOXWORKER_METRICNAMES_H

#include <string>

namespace buildboxworker {

struct MetricNames {
    static const std::string TIMER_NAME_DOWNLOAD_ACTION;
    static const std::string TIMER_NAME_UPDATE_BOTSESSION;
    static const std::string TIMER_NAME_CREATE_BOTSESSION;
    static const std::string TIMER_NAME_EXECUTE_ACTION;
    static const std::string GAUGE_NUM_JOBS_RUNNING_ON_WORKER;
    static const std::string GAUGE_MAXIMUM_WORKER_CAPACITY;
};

} // namespace buildboxworker

#endif // BUILDBOXWORKER_METRICNAMES_H
