// Copyright 2024 Bloomberg Finance L.P
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef INCLUDED_RUMBAD_FILEWRITER
#define INCLUDED_RUMBAD_FILEWRITER

#include <build/buildbox/local_execution.pb.h>

#include <rumbad_writer.h>

namespace rumbad {

class FileWriter : public Writer {
  public:
    FileWriter(const std::string &filepath);
    void write(const std::vector<Message> &messages) override;

  private:
    std::string d_logFilepath;
};

} // namespace rumbad

#endif
