// Copyright 2026 Bloomberg Finance L.P
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include <string>

namespace trexe {
namespace metricnames {
// Namespace containing all the strings under which metrics are published.

// Trexe main
const std::string RAISED_EXCEPTION = "raised_exception";
const std::string SUCCESS = "success";

} // namespace metricnames
} // namespace trexe
