/*
 * Copyright 2025 Bloomberg LP
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <buildboxcommon_protos.h>
#include <buildboxcommon_schedulingmetadata.h>

using namespace buildboxcommon;

const std::string SchedulingMetadataGenerator::HEADER_NAME =
    "build.buildgrid.schedulingmetadata-bin";

void SchedulingMetadataGenerator::attachSchedulingMetadata(
    grpc::ClientContext *context) const
{
    if (d_schedulingMetadata.ByteSizeLong() > 0) {
        std::string serialized_metadata;
        if (d_schedulingMetadata.SerializeToString(&serialized_metadata)) {
            context->AddMetadata(HEADER_NAME, serialized_metadata);
        }
    }
}
